/*
 *========================================================================
 * $Id: grade_answer.c 30 2004-01-25 22:38:21Z rgb $
 *
 * See copyright in copyright.h and the accompanying file COPYING
 *========================================================================
 */

/*
 *========================================================================
 * Problem presentation routines.  Presentation only -- timer is started
 * elsewhere, and then results are collected and evaluated still elsewhere.
 *========================================================================
 */

#include "flashcard.h"

void get_answer(int *pi)
{

 int c1,c2;

 if(verbose){
   wprintw(historyb.win,"get_answer: Entry.\n");
   wrefresh(historyb.win);
 }

 /*
  * This routine blocks until a valid answer is in answer.  It snoops and
  * processes all leading Esc-char control sequences.  It internally
  * replays the current problem if necessary on entry of a blank line.
  */

 answer[0] = (char)0;
 c1 = 27;  /* Gain entry into primary loop */
 while(answer[0] == (char)0 || c1 == 27){

   /*
    * We snoop the queue.  If the first character is 27/Esc, we get the
    * second (the command character) and take the appropriate action.  The
    * appropriate actions will be run a menu or execute a reconfiguration
    * command.  We loop on Esc's (by pairs) until we hit the first
    * character that isn't Esc.
    */
   noecho();
   c1 = wgetch(answerb.win);
   while(c1 == 27){
     c2 = wgetch(answerb.win);
     options(c2);
     if(state.suspend == YES) {
       echo();
       return;
     }
     present_problem(*pi);
     noecho();
     c1 = wgetch(answerb.win);
   }
   ungetch(c1);
   echo();
   if(state.suspend == YES){
     return;
   }

   answer[0] = (char)NULL;
   wgetnstr(answerb.win,answer,64);
   wrefresh(answerb.win);
  
   /*
    * If the answer we got was null (the empty string) we
    * we replay.
   */
   if(answer[0] == (char)0){
     wprintw(historyb.win,"-------------------------Replay-------------------------\n");
     present_problem(*pi);
   }

 }

 /*
  * We only got out of the loop if we got a non-escaped, non-null answer.
  * Here it is (for the record).
  */
 /*
  * Remove the answer entered and prompt right away so it doesn't block
  * the next problem.
  */
 mvwprintw(answerb.win,1,0,"");
 wclrtoeol(answerb.win);
 wrefresh(answerb.win);

 if(verbose){
   wprintw(historyb.win,"got_answer: got answer = %s\n",answer);
   wrefresh(historyb.win);
 }

}

void grade_answer(int *pi)
{

 /*
  * loop until we get an answer.  Don't worry, get_answer blocks until one
  * enters or types SOMETHING and returns 1 only when there is an answer
  * to be processed in answer.
  */
 get_answer(pi);
 if(state.suspend == YES) {
   /* Kill the timer so that the whole problem is "done" ungraded */
   pthread_cancel(timertid);
   return;
 }
 /*
  * In test mode, cancel the timer window right away once the answer is
  * collected.
  */
 if(state.mode == TEST) pthread_cancel(timertid);

 /*
  * Then determine whether or not the answer entered was correct.
  * All problem-type-specific magic is encapsulated in here.  They
  * return the result in "correctflag", although it might be better
  * still to make these routines integer functions and return 1 for
  * correct and 0 for incorrect.  They are actually "tallied" below.
  */
 switch(problem_list[*pi].type[0]){
   case 'A':
     grade_arithmetic(pi);
     break;
   case 'M':
     grade_multiple_choice(pi);
     break;
   case 'S':
     grade_spelling(pi);
     break;
   case 'T':
     grade_true_false(pi);
     break;
   default:
     /* Should never get here as we should have punted already above */
     wprintw(historyb.win,"Error:  Unknown problem type in %s.\n",state.flashfile);
     wrefresh(historyb.win);
     dump_problem(pi,0);
     return;
     break;
 }

 if(pstate.correct){
   if(state.mode == PRACTICE){
     wprintw(historyb.win,"%s is Right!\n",saveanswer);
     wprintw(historyb.win,"--------------------------------------------------------\n");
     wclrtoeol(historyb.win);
     wrefresh(historyb.win);
   }
   /*
    * This is always true in test mode, but only maybe true in practice mode
    * You don't get to count the practice repeats.
    */
   if(pstate.repeat_count == 0) increment_correct();
   if(state.mode == PRACTICE){
     if(pstate.repeat_count > 0) {
       pstate.repeat_count--;
       if(pstate.repeat_count == 0) wprintw(historyb.win,"\n");
     }
   }
 } else {
   if(pstate.repeat_count == 0) increment_incorrect();
   wprintw(historyb.win,"%s is Wrong!\n",saveanswer);
   if(state.mode == PRACTICE){
     /* explanation is guaranteed to neither begin nor end with a LF */
     wprintw(historyb.win,"=======================Explanation=======================\n");
     wprintw(historyb.win,"%s\n",problem_list[*pi].explanation);
     wprintw(historyb.win,"========================================================\n");
     if(state.repeat){
       if(state.repeat == 1)wprintw(historyb.win,"\nRepeat the correct answer >> %s << %d time!\n",problem_list[*pi].answer,state.repeat);
       if(state.repeat > 1) wprintw(historyb.win,"\nRepeat the correct answer >> %s << %d times!\n",problem_list[*pi].answer,state.repeat);
     }
     wclrtoeol(historyb.win);
     /* redo this problem */
     pstate.repeat_count = state.repeat;
   }
 }

 /*
  * In non-test mode cancel the timer window before leaving grade_answer.
  * We don't count problem presentation time, and a new timer will be
  * spawned after the next problem is presented.
  */
 if(state.mode != TEST) pthread_cancel(timertid);
 
 wrefresh(historyb.win);

 if(verbose){
   wprintw(historyb.win,"Graded problem %d.\n",pi);
   wrefresh(historyb.win);
 }

}


void grade_multiple_choice(int *pi)
{

 int i,j,k,inlist,numaids,answer_index,required_cnt;

 if(verbose){
   wprintw(historyb.win,"Multiple Choice: grading answer %s\n",answer);
   wrefresh(historyb.win);
 }

 /*
  * A "valid" answer is in answer.  So we check it.  First we shift it
  * to upper case.
  */
 i = 0;
 while(answer[i] != 0) {
   answer[i] = toupper(answer[i]);
   i++;
 }
 strncpy(saveanswer,answer,LINEBUF);
 number_of_answers = parse(answer,answerlist,SHOWMAX,LINEBUF);

 if(verbose){
   wprintw(historyb.win,"Found %d answers\n",number_of_answers);
   for(i=0;i<number_of_answers;i++){
     wprintw(historyb.win,"Answer token string %s\n",answerlist[i]);
   }
   wrefresh(historyb.win);
 }

 /*
  * Get a list of all the UNIQUE aid tokens.  These are all the aids that
  * MUST be present in one form or another for the answer to be correct.
  * Note that these tokens do NOT correspond to the answer tokens, which
  * have to be dereferenced against the choiceletter vector, converted
  * into an index, and finally compared.
  */
 numaids = 0;
 if(verbose){
   wprintw(historyb.win,"We will show %d choices\n",problem_list[*pi].show);
   wrefresh(historyb.win);
 }

 for(i=0;i<problem_list[*pi].show;i++){
   /* Skip quickly to an actual answer (non-null aid). */
   if(problem_list[*pi].choices[i].aid[0] != 0){
     if(verbose){
       wprintw(historyb.win,"Getting aid token list, i = %d\n",i);
       wprintw(historyb.win,"aid = %s\n",problem_list[*pi].choices[i].aid);
       wprintw(historyb.win,"text = %s\n",problem_list[*pi].choices[i].text);
       wrefresh(historyb.win);
     }
     /*
      * In here means we got one.  Now we have to see if we already
      * have it in the aidlist[].
      */
     inlist = 0;
     for(j=0;j<numaids;j++){
       if(strncmp(problem_list[*pi].choices[i].aid,problem_list[*pi].choices[j].aid,LINEBUF) == 0){
         if(verbose){
           wprintw(historyb.win,"%s already in list, numaids = %d\n",
              problem_list[*pi].choices[i].aid,numaids);
           wrefresh(historyb.win);
         }
         inlist = 1;
	 break;
       }
     }
     if(inlist == 0) {
       strncpy(aidlist[numaids++],problem_list[*pi].choices[i].aid,LINEBUF);
     }
   }
 }
 if(verbose){
   wprintw(historyb.win,"Found %d aids\n",numaids);
   for(i=0;i<numaids;i++){
     wprintw(historyb.win,"Actual unique aid string[%d]: %s\n",i,aidlist[i]);
   }
   wrefresh(historyb.win);
 }

 /*
  * So let's do it.  We turn each answer token into an index into
  * choicelist by subtracting 'A' (clever enough).  We look at the aid
  * for this entry.  If it is null (empty) we picked a wrong answer and
  * mark the problem wrong and skip the rest of the check.  We start
  * by assuming that the problem is right.
  */
 pstate.correct = YES;
 for(i=0;i<number_of_answers;i++){
   answer_index = (int)(answerlist[i][0]-'A');
   if(verbose){
     wprintw(historyb.win,"Answer index = %d, aid = %s\n",
       answer_index,problem_list[*pi].choices[choicelist[answer_index]].aid);
     wrefresh(historyb.win);
   }
   /* plain old wrong answer */
   if(problem_list[*pi].choices[choicelist[answer_index]].aid[0] == 0) {
     pstate.correct = NO;
     break;
   }
 }
 if(verbose){
   wprintw(historyb.win,"First pass (wrong answers) pstate.correct = %d\n",pstate.correct);
   wrefresh(historyb.win);
 }

 if(pstate.correct){
   /*
    * Now we run through the answerlist.  If the nonzero aid corresponding
    * to an answer is in the aidlist, we mark it OUT of the aidlist
    * and increment a counter.  If we find all distinct aid's at least
    * once, the counter should end up equalling numaids and the answer
    * is correct.  Otherwise we set correctflag to 0 as the answer is
    * missing at least one required component.
    */
   required_cnt = 0;
   for(i=0;i<number_of_answers;i++){
     answer_index = (int)(answerlist[i][0]-'A');
     if(verbose){
       wprintw(historyb.win,"Number of actual unique answers/aids = %d\n",numaids);
       wrefresh(historyb.win);
     }
     for(j=0;j<numaids;j++) {
       if(verbose){
         wprintw(historyb.win,"comparing answer[%d]=%s to aidlist[%d]=%s\n",
	   i,problem_list[*pi].choices[choicelist[answer_index]].aid,j,aidlist[j]);
         wrefresh(historyb.win);
       }
       if(strncmp(problem_list[*pi].choices[choicelist[answer_index]].aid,
                    aidlist[j],LINEBUF) == 0) {
         required_cnt++;
	 aidlist[j][0] = 0;
       }
       if(verbose){
         wprintw(historyb.win,"count = %d\n",required_cnt);
         wrefresh(historyb.win);
       }
     }
   }
   if(required_cnt != numaids) pstate.correct = NO;
 }
 if(verbose){
   wprintw(historyb.win,"second pass (all required answers) pstate.correct = %d\n",pstate.correct);
   wrefresh(historyb.win);
 }

}

void grade_arithmetic(int *pi)
{

 int correct_answer,my_answer;
 if(verbose){
   wprintw(historyb.win,"Arithmetic: collecting and grading %d from problem list\n",*pi);
   wrefresh(historyb.win);
 }

 /*
  * A "valid" answer is in answer.  Now we determine the correctness of
  * the answer entered.
  */
 correct_answer = atoi(problem_list[*pi].answer);
 my_answer = atoi(answer);
 if(verbose){
   wprintw(historyb.win,"Got answer = %d, correct answer = %d\n",my_answer,correct_answer);
  wrefresh(historyb.win);
 }

 if(correct_answer == my_answer) {
   pstate.correct = YES;
 } else {
   pstate.correct = NO;
 }

}

void grade_spelling(int *pi)
{

 int i,j,k,inlist,numaids,answer_index,required_cnt;

 if(verbose){
   wprintw(historyb.win,"Spelling: collecting and grading %d from problem list\n",*pi);
   wrefresh(historyb.win);
 }

 /*
  * A "valid" answer is in answer.  We determine if the entered answer is
  * correct.
  */
 strncpy(saveanswer,answer,LINEBUF);
 /*
  * parse kills the original string so we cannot use it later
  * to check multiword answers if we parse it now.  Note that
  * AT THE MOMENT multiword answers can only occur in the
  * answer tag because I don't know how to easily check for separators
  * in the tty flashcard answerbox.  Eventually a web/gui interface will
  * have multiple boxes, and eventually I'll work out something for
  * tty.  For the moment, it looks like we're going to need to keep
  * the <answer> tag for spelling even though it is ALMOST redundant
  * with using a single <choice>.  A single <choice> won't manage
  * multiword spelling.
  *
  * Don't checking answers just byte?
  */
 if(problem_list[*pi].answer[0] == 0){
   number_of_answers = parse(answer,answerlist,SHOWMAX,LINEBUF);
   if(verbose){
     wprintw(historyb.win,"Found %d answers\n",number_of_answers);
     for(i=0;i<number_of_answers;i++){
       wprintw(historyb.win,"Answer token string %s\n",answerlist[i]);
     }
     wrefresh(historyb.win);
   }
 } else {
   if(verbose){
     wprintw(historyb.win,"Grading answer string %s\n",saveanswer);
     wrefresh(historyb.win);
   }
 }

 /*
  * We now do a split.  In 99% of all spelling problems, there will be
  * only one correct answer, and the <answer> tag will suffice.
  * In an effort to keep from breaking things all at once, let's try
  * to preserve <answer> as an acceptable tag for spelling problems
  * for the moment, superceding choices if both are present.
  */
 if(problem_list[*pi].answer[0] == 0){

   /*
    * Get a list of all the UNIQUE aid tokens.  These are all the aids that
    * MUST be present in one form or another for the answer to be correct.
    * Note that these tokens do NOT correspond to the answer tokens, which
    * have to be dereferenced against the choiceletter vector, converted
    * into an index, and finally compared.
    */
   numaids = 0;
   for(i=0;i<problem_list[*pi].numchoices;i++){
     /*
      * There should be nothing but actual answers here. Unlike
      * multiple choice, if there is no aid string, the answer is
      * assumed to be a simple OR list of homonyms.  So we still have
      * to compute unique aid's but NULL now counts as an aid so we
      * don't skip it.
      */
     if(verbose){
       wprintw(historyb.win,"Getting aid token list, i = %d\n",i);
       wprintw(historyb.win,"aid = %s\n",problem_list[*pi].choices[i].aid);
       wprintw(historyb.win,"text = %s\n",problem_list[*pi].choices[i].text);
       wrefresh(historyb.win);
     }

     /*
      * We check to see if we already have the current aid (if any) in
      * the aidlist[].
      */
     inlist = 0;
     for(j=0;j<numaids;j++){
       if(strncmp(problem_list[*pi].choices[i].aid,problem_list[*pi].choices[j].aid,LINEBUF) == 0){
         if(verbose){
           wprintw(historyb.win,"%s already in list, numaids = %d\n",
              problem_list[*pi].choices[i].aid,numaids);
           wrefresh(historyb.win);
         }
         inlist = 1;
         break;
       }
     }
     if(inlist == 0) {
       strncpy(aidlist[numaids++],problem_list[*pi].choices[i].aid,LINEBUF);
     }
   }
   if(verbose){
     wprintw(historyb.win,"Found %d aids\n",numaids);
     for(i=0;i<numaids;i++){
       wprintw(historyb.win,"Actual unique aid string[%d]: %s\n",i,aidlist[i]);
     }
     wrefresh(historyb.win);
   }

   /*
    * Now we run through the answerlist and compare EACH answer entered
    * to EACH string in choices.  We look for two things:
    *  a) A string in the answerlist that doesn't correspond to any
    * choice is instantly wrong and break out of the rest of checking
    * immediately.
    *  b) Each string that DOES match a string in the answerlist causes
    * that string's aid to be marked out of the aidlist and a counter
    * to be incremented.  If we find all the distinct aid's at least
    * once, the counter should equal numaids and the answer is correct.
    * Otherwise, it is incorrect as a "required" spelling is missing
    * at least one example thereof.
    *
    * A warning.  We can no longer mark out the aidlist with a null,
    * because null is a default aid that counts.  So we use 0x01, which
    * cannot be entered in a text file as a possible aid and hence is
    * safe.
    */
   required_cnt = 0;
   pstate.correct = YES;
   for(i=0;i<number_of_answers;i++){
     inlist = 0;
     j = 0;
     while(inlist == 0 && j < problem_list[*pi].numchoices){
       if(strncmp(answerlist[i],problem_list[*pi].choices[j].text,LINEBUF) == 0) {
         /*
          * We found the entered answer in the list of answers!
          * Now we have to mark out its aid.  We now the aid as
	  * problem_list[*pi].choices[j].aid.  If this matches an
	  * aid in aidlist, we set the aidlist entry to 1 (number,
	  * not ascii '1') to mark it out of further testing and
	  * increment required_cnt.
          */
        if(verbose){
          wprintw(historyb.win,"Entered answer %s matches correct answer %s\n",answerlist[i],problem_list[*pi].choices[j].text);
          wrefresh(historyb.win);
         }
         inlist = 1;
         for(k=0;k<numaids;k++) {
           if(verbose){
             wprintw(historyb.win,"comparing answer[%d]=%s to aidlist[%d]=%s\n",
                j,problem_list[*pi].choices[j].aid,k,aidlist[k]);
             wrefresh(historyb.win);
           }
           if(strncmp(problem_list[*pi].choices[j].aid,aidlist[k],LINEBUF) == 0) {
             required_cnt++;
             aidlist[k][0] = 1;
           }
           if(verbose){
             wprintw(historyb.win,"count = %d\n",required_cnt);
             wrefresh(historyb.win);
           }
	 }
       }
       j++;
     }
     if(inlist == 0){
       if(verbose){
         wprintw(historyb.win,"Answer %s is incorrect!\n",answerlist[i]);
         wrefresh(historyb.win);
       }
       pstate.correct = NO;
       break;
     }
   }
   if(required_cnt != numaids) pstate.correct = NO;
 } else {
    if(strncmp(answer,problem_list[*pi].answer,64)!=0) {
      pstate.correct = NO;
    } else {
      pstate.correct = YES;
    }
   if(verbose){
     wprintw(historyb.win,"answer defined, string %s yields pstate.correct = %d\n",answer,pstate.correct);
     wrefresh(historyb.win);
   }

 }

}


void grade_true_false(int *pi)
{

 char answerchar;

 if(verbose){
   wprintw(historyb.win,"True/False: grading answer %s.\n",answer);
   wrefresh(historyb.win);
 }

 /*
  * A "valid" answer is in answer.  We now determine whether or not the
  * answer entered is correct.
  */

 /*
  * ignore everything but the first character, upshifting it for
  * consistency (so 't' and 'T' both work).
  */
 if((char)toupper(problem_list[*pi].answer[0]) == (char)toupper(answer[0])) {
   pstate.correct = YES;
 } else {
   pstate.correct = NO;
 }

}

void increment_correct()
{

   state.correct++;
   state.count++;
   show_score();
}

void increment_incorrect()
{

   state.incorrect++;
   state.count++;
   show_score();

}

