/*
 *========================================================================
 * $Id: parsexml_flashfile.c 27 2004-01-25 06:35:11Z rgb $
 *
 * See copyright in copyright.h and the accompanying file COPYING
 *========================================================================
 */

#include "flashcard.h"

/*
 *========================================================================
 * parsexml()
 *========================================================================
 */

 /*
  * If state.state is LOAD, the file stats, so it is time to actually
  * load it and parse it.  If loading fails BEFORE any records are parsed
  * (so the old problem_list is still valid) and
  * state.suspend is YES, state.state should go to RUN to resume.  If
  * parsing fails but the old problem list is corrupted, state.state
  * has to go back to NONE (with an Error Message!) because we can
  * no longer resume the old one.  If parsing succeeds, a new file
  * is loaded and the old problem set is no longer even there.  At that
  * point state.state goes to RUN if state.suspend is NO (so there is
  * no old session to terminate and wrap up) and END if state.suspend is
  * YES (so the old session can print out final statistics, interact
  * with any possible future workminder tool, etc.)
  */

void parsexml_flashfile()
{

 /* Generic loop indices, and node limit */
 int i,j,k,numnodes,delay = 0;

 /* xml (tree) objects */
 xmlDocPtr problem_doc;
 xmlNodePtr cur = NULL,problem = NULL,root = NULL,choicecur = NULL;
 /* xml (path) objects) */
 xmlXPathContextPtr xp_problem_doc = NULL;
 xmlXPathObjectPtr xp_op = NULL;
 /* a pointer to an attribute value */
 char *attribute_value;

 /*
  * This is the running number of problems read in so far.  We have to
  * null the problem_list ptr so that realloc can function like malloc on
  * the first call.
  */
 number_of_problems = 0;
 problem_list = NULL;

 if(verbose){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Preparing to read in and parse file %s\n",state.flashfile);
   wrefresh(historyb.win);
   sleep(delay);
 }

 problem_doc = xmlParseFile(state.flashfile);
 /* in case something went wrong with read or parse */
 if(problem_doc == NULL){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Error! Couldn't xmlParseFile(%s).  Is this a flashfile?\n",state.flashfile);
   wrefresh(historyb.win);
   if(state.suspend == YES) {
     state.state = RUN;
   } else {
     state.state = NONE;
   }
   return;
 } else {
   root = xmlDocGetRootElement(problem_doc);
   /* if there is no root element */   
   if(root == NULL) {
     wprintw(historyb.win,"parsexml_flashfile():\n");
     wprintw(historyb.win,"Error! No document root in %s. Is this a flashfile?\n",state.flashfile);
     wrefresh(historyb.win);
     if(state.suspend == YES) {
       state.state = RUN;
     } else {
       state.state = NONE;
     }
     return;
   } 
   /* if there is no "flashcard" root element */   
   if( strcasecmp(root->name,"flashcard") != 0 ) {
     wprintw(historyb.win,"parsexml_flashfile():\n");
     wprintw(historyb.win,"Error! No <flashcard> tag in %s.  Is this a flashfile?\n",state.flashfile);
     wrefresh(historyb.win);
     if(state.suspend == YES) {
       state.state = RUN;
     } else {
       state.state = NONE;
     }
     return;
   }
 }

 if(verbose){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Successfully read in file %s.  Creating Xpath ctxt.\n",state.flashfile);
   wrefresh(historyb.win);
   sleep(delay);
 }

 xp_problem_doc = xmlXPathNewContext(problem_doc);

 /*
  * Time to parse out the elements and put them away.
  */

 /*
  * Getting /flashcard so we can extract the presentation attribute.
  */
 if(verbose){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Got Xpath context. About to evaluate the /flashcard xpath.\n");
   wrefresh(historyb.win);
   sleep(delay);
 }
 i = 0;
 xp_op = xmlXPathEval((const xmlChar *)"/flashcard",xp_problem_doc);
 /* Did we get an xpath object at all for this path? */
 if( xp_op == NULL ) {
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Error! xpath cannot parse %s!\n",state.flashfile);
   wrefresh(historyb.win);
   if(state.suspend == YES) {
     state.state = RUN;
   } else {
     state.state = NONE;
   }
   return;
 }
 /* If we did, is it a nodeset? */
 if( xp_op->type != XPATH_NODESET ){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Error! No nodes found by xpath in hostfile!\n");
   wrefresh(historyb.win);
   if(state.suspend == YES) {
     state.state = RUN;
   } else {
     state.state = NONE;
   }
   return;
 }
 if(verbose){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Should have an xpath object pointer to <flashcard>.\n");
   wprintw(historyb.win,"xp_op = %0x\n",xp_op);
   wrefresh(historyb.win);
   sleep(delay);
 }
 /* There has better be exactly one of these... */
 cur = xp_op->nodesetval->nodeTab[0];
 if(verbose){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"cur = %0x\n",cur);
   wrefresh(historyb.win);
   sleep(delay);
 }

 /*
  * MUST free attribute_value when done.
  */
 attribute_value = (char *)xmlGetProp(cur, (const xmlChar *) "presentation");
 if(state.presentation == UNSET){
   if(verbose){
     wprintw(historyb.win,"parsexml_flashfile():\n");
     wprintw(historyb.win,"No state.presentation.  Looking at attribute %s.\n",attribute_value);
     wrefresh(historyb.win);
     sleep(delay);
   }
   if(attribute_value[0]){
     if(attribute_value[0] == 'l' || attribute_value[0] == 'L'){
       state.presentation = LIST;
       if(verbose){
         wprintw(historyb.win,"parsexml_flashfile():\n");
         wprintw(historyb.win,"LIST presentation\n");
         wrefresh(historyb.win);
         sleep(delay);
       }
     } else if(attribute_value[0] == 'r' || attribute_value[0] == 'R'){
       state.presentation = RANDOM;
       if(verbose){
         wprintw(historyb.win,"parsexml_flashfile():\n");
         wprintw(historyb.win,"RANDOM presentation\n");
         wrefresh(historyb.win);
         sleep(delay);
       }
     } else {
       wprintw(historyb.win,"parsexml_flashfile():\n");
       wprintw(historyb.win,"Warning!  Presentation type %s not recognized!\n",attribute_value);
       wprintw(historyb.win,"Using default LIST presentation.\n");
       wrefresh(historyb.win);
       state.presentation = LIST;
     }
   } else {
     /* The default */
     state.presentation = LIST;
     if(verbose){
       wprintw(historyb.win,"parsexml_flashfile():\n");
       wprintw(historyb.win,"Using default LIST presentation\n");
       wrefresh(historyb.win);
       sleep(delay);
     }
   }
 } else {
   switch(state.presentation){
     case LIST:
       if(verbose){
         wprintw(historyb.win,"parsexml_flashfile():\n");
         wprintw(historyb.win,"Using LIST presentation set at command line.\n");
         wrefresh(historyb.win);
         sleep(delay);
       }
       break;
     case RANDOM:
       if(verbose){
         wprintw(historyb.win,"parsexml_flashfile():\n");
         wprintw(historyb.win,"Using RANDOM presentation set at command line.\n");
         wrefresh(historyb.win);
         sleep(delay);
       }
       break;
     default:
       wprintw(historyb.win,"parsexml_flashfile():\n");
       wprintw(historyb.win,"Error!  Presentation type not known.  This cannot happen!\n");
       wprintw(historyb.win,"^C out and fix or report the problem.\n");
       wrefresh(historyb.win);
       beep();
       beep();
       beep();
       sleep(100);
       break;
   }
 }
 xmlFree(attribute_value);

 /*
  * If we get here, any suspended session is now null and void as
  * its old state.presentation has just been overwritten.  So we must
  * FREE any allocated problem_list and any other resources and
  * set state.state to END so that a secondary routine can finish
  * cleaning up and tallying things, IF the rest of the routine
  * successfully parses the current file.
  */
 if(state.suspend == YES){
   if(verbose){
     wprintw(historyb.win,"parsexml_flashfile():  Freeing iorder and problem_list.\n");
     wrefresh(historyb.win);
     sleep(delay);
   }
   if(iorder) free(iorder);
   if(problem_list) free(problem_list);
   state.state == END;
 }

 /*
  *==================================================================
  * /flashcard/repeat
  *
  * suggested repeat tag, can be incremented or decremented by user.
  *==================================================================
  
  */
 if(verbose){
   wprintw(historyb.win,"About to evaluate the /flashcard/repeat xpath.\n");
   wrefresh(historyb.win);
   sleep(delay);
 }
 i = 0;
 xp_op = xmlXPathEval((const xmlChar *)"/flashcard/repeat",xp_problem_doc);
 /*
  * Did we get an xpath object at all for this path?  Is it a nodeset?
  * Are there any nodes?
  */
 if( xp_op != NULL) {
   if(xp_op->type == XPATH_NODESET ) {
     numnodes = xp_op->nodesetval->nodeNr;
     if(numnodes){
       cur = xp_op->nodesetval->nodeTab[0]->xmlChildrenNode;
       state.repeat = atoi((char *)cur->content);
       if(verbose){
         wprintw(historyb.win,"parsexml_flashfile(): repeat is set to %d\n",state.repeat);
         wrefresh(historyb.win);
         sleep(delay);
       }
     }
   }
 }
 xmlXPathFreeObject(xp_op);

 /*
  *==================================================================
  * /flashcard/problem
  *
  * Note well that there are several distinct types of problems, but
  * they all share a single struct for storage so that they can be
  * indiscriminately mixed in a flash file.  We DO need to create
  * quite separate parsing branches for each file type.  We also
  * should start to use reusable subroutines for parsing out particular
  * kinds of values within the problem descriptors.
  *==================================================================
  */
 if(verbose){
   wprintw(historyb.win,"About to evaluate the /flashcard/problem xpath.\n");
   wrefresh(historyb.win);
   sleep(delay);
 }
 i = 0;
 xp_op = xmlXPathEval((const xmlChar *)"/flashcard/problem",xp_problem_doc);
 /* Did we get an xpath object at all for this path? */
 if( xp_op == NULL ) {
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Error! xpath cannot parse any <problem>s.  Is this a flashfile?\n");
   wrefresh(historyb.win);
   state.state = NONE;
   return;
 }
 /* If we did, is it a nodeset? */
 if( xp_op->type != XPATH_NODESET ){
   wprintw(historyb.win,"parsexml_flashfile():\n");
   wprintw(historyb.win,"Error! no problem nodes found by xpath.  Is this a flashfile?\n");
   wrefresh(historyb.win);
   state.state = NONE;
   return;
 }

 if(verbose){
   wprintw(historyb.win,"Should have an xpath object pointer to a problem.\n");
   wrefresh(historyb.win);
   sleep(delay);
 }

 /*
  * xp_op now contains an xpath object pointer to the /flashpath/problem
  * list in no particular order.
  *
  * Time to walk the dog-ument and pull out the contents of each problem.
  * This sets cur to point to each <flashpath><problem> in turn:
  * The loop puts the results into an array, allocating the space required
  * as each element is added.
  */
 numnodes = xp_op->nodesetval->nodeNr;

 if(verbose){
   wprintw(historyb.win,"There are %d problem nodes.  We walk them.\n",numnodes);
   wrefresh(historyb.win);
   sleep(delay);
 }

 for(i = 0 ; i < numnodes; i++){
   /* count the problems so far */
   number_of_problems++;
   /* allocate space for the current problem */
   problem_list = realloc(problem_list,(number_of_problems*sizeof(Problem)));
   problem_list[i].choices = (Choice *)NULL;
   /* prepare to parse the contents of THIS problem */
   problem = cur = xp_op->nodesetval->nodeTab[i]->xmlChildrenNode;
   if(verbose){
     wprintw(historyb.win,"Extracting problems from point = %x\n",cur);
     wrefresh(historyb.win);
     sleep(delay);
   }
   while (cur != NULL) {
     /*
      * Extract the "type" tag contents.  Upshift it to caps.
      */
     if( xmlStrcmp(cur->name, (const xmlChar *) "type") == 0) {
       strncpy(problem_list[i].type,(char *)cur->xmlChildrenNode->content,TYPEBUF);
       problem_list[i].type[0] = (char)toupper(problem_list[i].type[0]);
       problem_list[i].type[1] = 0;
       if(verbose){
         wprintw(historyb.win,"Extracting problem of type %s at point = %x\n",problem_list[i].type,problem);
         wrefresh(historyb.win);
       }
       /*
        * These are some defaults for the new problem, by type.
	*/
       switch(problem_list[i].type[0]){
         case 'A':
           strncpy(problem_list[i].prompt,"Enter numerical answer:",LINEBUF);
	   break;
         case 'S':
           strncpy(problem_list[i].prompt,"Please spell word correctly:",LINEBUF);
	   break;
	 case 'M':
           strncpy(problem_list[i].prompt,"Select all correct answer(s):",LINEBUF);
	   break;
	 case 'T':
           strncpy(problem_list[i].prompt,"T(rue) or F(alse)?:",LINEBUF);
	   break;
	 default:
           strncpy(problem_list[i].prompt,"No prompt:",LINEBUF);
	   break;
       }
       parse_problem(problem,i);
     }

     cur = cur->next;
   }
 }
 xmlXPathFreeObject(xp_op);

 /*
  * When we have finished reading in the list of problems, we need to
  * allocate room for the shuffle vector.
  */
 iorder = (int *)malloc(number_of_problems*sizeof(int));

 if(verbose){
   wprintw(historyb.win,"Finished parsing the problem doc.\n");
   wrefresh(historyb.win);
   sleep(delay);
 }

 /*
  * We are done (we hope) with the xml
  * structs and don't want to needlessly leak memory.
  */
 xmlXPathFreeContext(xp_problem_doc);
 xmlFreeDoc(problem_doc);

 /*
  * As the last thing we do, if state.suspend is YES set state.state
  * to END to remind the work loop to clean up.  Otherwise set it to
  * RUN to run the shiny new problem set.  Since the file is valid
  * and about to run either way, we save its name for later presentation
  * in end_flashfile().
  */
 if(state.suspend == YES) {
   state.state = END;
   if(verbose){
     wprintw(historyb.win,"parsexml_flashfile():.\n");
     wprintw(historyb.win,"state.suspend is YES, state.state is END.\n");
     wrefresh(historyb.win);
     sleep(delay);
   }
 } else {
   state.state = RUN;
   strncpy(flashfile_save,state.flashfile,LONGBUF);
   if(verbose){
     wprintw(historyb.win,"parsexml_flashfile():.\n");
     wprintw(historyb.win,"state.suspend is NO, state.state is RUN.\n");
     wrefresh(historyb.win);
     sleep(delay);
   }
 }


}

/*
 * One size FINALLY fits all.  Parse all problem content.
 */
void parse_problem(xmlNodePtr problem,int i)
{

 int textsize,aidsize,nshow,nchoices = 0,delay=0;
 char text[1024],*aid = 0;

 /*
  * A variety of things need to be nulled so that they show as being
  * uninitialized later.  This is IMPORTANT!
  */
 problem_list[i].show = 0;
 problem_list[i].audio[0] = 0;
 problem_list[i].answer[0] = 0;
 problem_list[i].question[0] = 0;
 problem_list[i].explanation[0] = 0;

 /*
  * Most fields get fix_lf run on them, because we don't want
  * leading lf's or trailing blanks or tabs or lf's as they
  * mess up our presentation format.
  */
 while (problem != NULL) {

   /*
    * Extract the "audio" tag contents.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "audio") == 0) {
     strncpy(problem_list[i].audio,(char *)problem->xmlChildrenNode->content,LONGBUF);
     fix_lf(problem_list[i].audio,LONGBUF);
   }

   /*
    * Extract the "question" tag contents.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "question") == 0) {
     strncpy(problem_list[i].question,(char *)problem->xmlChildrenNode->content,LONGBUF);
     fix_lf(problem_list[i].question,LONGBUF);
   }

   /*
    * Extract the "prompt" tag contents.  It has a default value above
    * that will suffice in most cases.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "prompt") == 0) {
     strncpy(problem_list[i].prompt,(char *)problem->xmlChildrenNode->content,LINEBUF);
   }

   /*
    * Extract the "answer" tag contents.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "answer") == 0) {
     strncpy(problem_list[i].answer,(char *)problem->xmlChildrenNode->content,LONGBUF);
     fix_lf(problem_list[i].answer,LONGBUF);
   }

   /*
    * Extract the "explanation" tag contents.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "explanation") == 0) {
     strncpy(problem_list[i].explanation,(char *)problem->xmlChildrenNode->content,PAGEBUF);
     fix_lf(problem_list[i].explanation,PAGEBUF);
   }

   /*
    * Extract the "show" tag contents.  It is an integer number, we hope.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "show") == 0) {
     problem_list[i].show = atoi((char *)problem->xmlChildrenNode->content);
   }

   /*
    * Here we parse out all the choices (and there will be
    * more than one, in general).  We also reallocate just enough space
    * as they appear.  Note that choices go into a struct of their own.
    *
    * Curiously enough, we use the aid (answer id) attribute of the choice
    * to encode the answer pattern remarkably powerfully given the
    * simplicity of the algorithm.  All choices with non-null aid strings
    * are answers as opposed to foolers.  All choices that have the SAME
    * aid string are evaluated with an implicit "or" (don't worry, they'll
    * have distinct choice letters in the random presentation).  All
    * choices with DISTINCT aid strings are evaluated with an implicit
    * "and" and must ALL be selected in the answer.
    *
    * This won't do arbitrary logic, sorry, but it will certainly do type
    * K questions as well as things like (note parentheses, which are
    * implicit in the algorithm):
    *   A or B  (A,B,AB correct)
    *   (A or B) and C (AC,BC,ABC)
    *   (A or B) and (C or D) (AC,AD,BC,BD,ABC,ABD,ACD,BCD,ABCD)
    *   A or B or C (A,B,C,AB,AC,BC,ABC)
    *   A and B and C (ABC)
    *
    * Can't do NOT logic except that all of these have an implicit
    * "and NOT any foolers".  Also cannot do exclusive OR, in case some
    * sadistic human can figure out how to give a question where the
    * correct answer is A OR B but both together aren't also correct.
    * Still, as I said, it should MORE than suffice for the range of
    * answer combinations likely to be needed to facilitate learning
    * and practice even up to the MCAT/SAT level, if I fancy up the
    * problem presentation engine to use e.g. latex2html generated
    * graphics.
    */
   if( xmlStrcmp(problem->name, (const xmlChar *) "choice") == 0) {

     /*
      * We get value and attribute at the beginning to facilitate handling.
      * Note that we have to xmlFree(aid) below or leak.
      */
     aid = (char *)xmlGetProp(problem, (const xmlChar *) "id");
     if(verbose){
       wprintw(historyb.win,"aid[%d] = %s\n",nchoices,aid);
       wrefresh(historyb.win);
       sleep(delay);
     }
     strncpy(text,(char *)problem->xmlChildrenNode->content,LONGBUF);
     fix_lf(text,LONGBUF);
     if(verbose){
       wprintw(historyb.win,"text[%d] = %s",nchoices,text);
       wrefresh(historyb.win);
       sleep(delay);
     }
     /* need sizes to control allocations */
     textsize = (int) strlen(text)+1;
     if(aid){
       aidsize = (int) strlen(aid)+1;
     } else {
       aidsize = 1;
     }

     /*
      * Allocate space for the current problem.  First we increment the
      * size of problem_list[i].choices by one Choice.  Then we allocate a
      * character string just large enough to hold the choice aid and text
      * strings at the pointer.  Finally, we fill them in.
      */

     problem_list[i].choices = (Choice *)realloc(problem_list[i].choices,(nchoices+1)*sizeof(Choice));
     problem_list[i].choices[nchoices].aid = (char *)malloc(aidsize*sizeof(char));
     problem_list[i].choices[nchoices].text = (char *)malloc(textsize*sizeof(char));
     if(aid){
       strncpy(problem_list[i].choices[nchoices].aid,aid,aidsize+1);
     } else {
       problem_list[i].choices[nchoices].aid[0] = (char)0;
     }
     strncpy(problem_list[i].choices[nchoices].text,text,textsize);

     /* count the choices so far */
     nchoices++;
     xmlFree(aid);

   }

   problem = problem->next;

 }

 /*
  * for many problem types, this will usually be zero.  If it is NOT
  * zero it is probably because it is either a multiple choice problem
  * or because we want to accept more than one answer, e.g.
  * through and threw, or to, too and two in spelling.  We may make
  * these presentation/evaluation decisions on a per-type basis, as
  * the problem GRADER is very much type dependent.
  */
 problem_list[i].numchoices = nchoices;

 /*
  * default is to show all the choices unless the problem has
  * no specified number.  We need to ensure later that we present
  * at least all the answers!
  */
 if(problem_list[i].show == 0) problem_list[i].show = nchoices;

 /*
  * Dump the newly parsed problem for a check, if necessary.  The
  * second argument is a delay to keep it from going by too fast.
  * In seconds.
  */
 if(verbose) dump_problem(i,0);

}

void fix_lf(char *buf,int buflen)
{

 int i,len,delay=0;

 /*
  * We strip all trailing LF's and blanks, as well as all leading LF's.
  * An "explanation" is everything in between.
  */
 len = i = 0;
 if(0){
   wprintw(historyb.win,"fix_lf: unstripped buf =:\n%s\n",buf);
   wrefresh(historyb.win);
 }
 while(buf[len] == 10) {len++;}
 /*
  * len now points to the first character that isn't a LF.  We shift
  * the whole string left this number of slots
  */
 while(buf[len] != 0 && len < buflen){
   buf[i++] = buf[len++];
 }
 /*
  * len now points at the first character that is null (string terminator)
  * OR is equal to buflen OR is equal to zero (null string).  We work our
  * way backwards, killing tab, LF, CR and space until we hit a character
  * that is none of these.
  */
 if(len){
   len--;
   while(buf[len] == 9 || buf[len] == 10 || buf[len] == 13 || buf[len] == 32) buf[len--] = 0;
 }
 if(0){
   wprintw(historyb.win,"fix_lf: stripped buf =:\n%s\n",buf);
   wrefresh(historyb.win);
   sleep(delay);
 }

}

