/*
 *========================================================================
 * $Id: present_utils.c 2 2004-01-16 16:35:02Z rgb $
 *
 * See copyright in copyright.h and the accompanying file COPYING
 *========================================================================
 */

/*
 *========================================================================
 * Problem presentation routines.  Presentation only -- timer is started
 * elsewhere, and then results are collected and evaluated still elsewhere.
 *========================================================================
 */

#include "flashcard.h"

int present_question(int pi)
{

 if(verbose){
   wprintw(historyb.win,"present_question(%d) entry.\n",pi);
   wrefresh(historyb.win);
 }

 /*
  * We start with a sanity check.  A problem must have a question in it.
  * the question content can be either an audio file and/or a question.
  * Both are nulled with the question is created, at least one must be
  * non-null now.
  */
 if(problem_list[pi].audio[0] == 0 && problem_list[pi].question[0] == 0){
   beep();
   wprintw(historyb.win,"Error! Problem %d has no presentation content!\n",pi);
   wprintw(historyb.win,"  Make sure every problem has either a question tag\n");
   wprintw(historyb.win,"  or an audio tag or both.\n\n");
   wrefresh(historyb.win);
   sleep(3);
   return(0);  /* failure */
 }

 /*
  * If an audio has been defined, we play it.  This supports the use
  * of the tool by the blind, by the way.
  */
 if(problem_list[pi].audio[0]){
   play_audio(pi);
 }

 /*
  * If a question has been defined we present it.  It is guaranteed
  * not to end in an LF, so we have to add one if we need one.
  */
 if(problem_list[pi].question[0]){
   if(problem_list[pi].type[0] == 'A'){
     wprintw(historyb.win,"%s ",problem_list[pi].question);
   } else {
     wprintw(historyb.win,"%s\n",problem_list[pi].question);
   }
   wrefresh(historyb.win);
 }

 if(verbose){
   wprintw(historyb.win,"present_question returning success.\n");
   wrefresh(historyb.win);
 }

 return(1);  /* success */

}

void present_choices(int pi)
{

 int i,j,itmp,*mlist;
 double drand;

 /*
  * We need to shuffle the choices and present them with e.g.
  * A,B,C,D in front.  To grade the result, we'll need a crossreference
  * array containing the indices of the problems, as we have to check
  * aid and work out a logic problem.  However, I'm now confident that
  * this will be pretty easy.  For the moment, we'll just clean up the
  * basic display.
  */

 /*
  * First we make a shuffled list of ALL choice indices.
  */
 for(i=0;i<problem_list[pi].numchoices;i++) choicelist[i] = i;
 shufflelist(choicelist,problem_list[pi].numchoices);

 /*
  * Second we have to ensure that the solution(s) are all in the
  * presentation list.
  */
 if(verbose){
   wprintw(historyb.win,"Show %d choices.\n",problem_list[pi].show);
   wrefresh(historyb.win);
 }
 for(i=problem_list[pi].show;i<problem_list[pi].numchoices;i++){
   /*
    * Swap them indices of each answer in the leftovers with a non-answer
    * in the presentation part of the list.  This is introduces bias, BTW,
    * so we can't JUST swap them into the presentation list.
    */
   while(problem_list[pi].choices[choicelist[i]].aid[0]) {
     /*
      * Pick a random j from the presentation part of the list and
      * if it isn't a solution component swap it down.
      */
     drand = 1.0*rand()/RAND_MAX;
     j = (int)problem_list[pi].show*drand;
     if(problem_list[pi].choices[choicelist[j]].aid[0] == 0){
       itmp = choicelist[j];
       choicelist[j] = choicelist[i];
       choicelist[i] = itmp;
     }
   }
 }
 /*
  * The presentation list should now contain all the answers and
  * random non-answers, although there is definitely a bias in
  * the distribution of the one relative to the other.  So we remove
  * the bias with a second shuffle, only of the presentation list.
  */
 shufflelist(choicelist,problem_list[pi].show);

 
 /*
  * Present the shuffled choices, using choiceletter[j] as the answer
  * key. Indent choices slightly.  Choices are guaranteed to neither
  * begin nor end with an LF, no matter how they were tagged.  This
  * is so we can get a uniform presentation.
  */
 for(j=0;j<problem_list[pi].show;j++){
   wprintw(historyb.win,"   %s: %s\n",
       choiceletter[j],problem_list[pi].choices[choicelist[j]].text);
 }
 wrefresh(historyb.win);

}

void present_prompt(char *prompt)
{

 /*
  * We do a check to be sure that the prompt "exists", but
  * it is pretty rudimentary.  The prompt (only) goes in the
  * "answer window".  If there is no prompt we put up a totally
  * generic prompt.
  */
 if(prompt != 0 && prompt[0] != 0){
   mvwprintw(answerb.win,1,0," %s ",prompt);
 } else {
   mvwprintw(answerb.win,1,0," Answer: ",prompt);
 }
 wclrtoeol(answerb.win);
 wrefresh(answerb.win);

}

void play_word(int pi)
{

 char cline[1024];

 wprintw(historyb.win,"Putting audio paths in <question> tags deprecated.\n");
 wprintw(historyb.win,"Please convert these all to <audio> tags.\n"); 
 wrefresh(historyb.win);

 snprintf(cline,1024,"/usr/bin/play %s",problem_list[pi].question);
 system(cline);

}

/*
 * Eventually we need to make this MUCH smarter and robust.  Particular
 * clever ideas include actually testing the path for existence with
 * a stat and punting if it doesn't, testing the magic of the file for
 * audio and punting if it isn't, and using a play library call instead
 * of a userspace/shell program to do the actual playing.  We also need
 * some ritual to follow if an error is encountered.
 */
void play_audio(int pi)
{

 char cline[1024];

 snprintf(cline,1024,"/usr/bin/play %s",problem_list[pi].audio);
 system(cline);

}

